# TODO Implement support for transparent redirects.
# Following redirection is implemented in HTTPClient but does
# not work for asynchronous requests. Also there is no support
# for POST redirects which may be also needed.

require 'httpclient'
require 'uu_os_connection/version'
require 'uu/os/http/http_get'
require 'uu/os/http/http_post'
require 'uu/os/http/http_method'
require 'uu/os/http/http_query'
require 'uu/os/http/http_headers'
require 'uu/os/http/http_request'
require 'uu/os/http/http_response'
require 'uu/os/util/url_converter'

# Patch allowing to stream request data using IO.pipe or STDIN,
# and also to use response stream directly for another request.
# Without this patch we will be forced to use temporary files.
# Based on https://github.com/nahi/httpclient/issues/112
# @private
module HTTP # :nodoc:
  # @private
  class Message # :nodoc:
    # @private
    class Body # :nodoc:
      # @private
      def remember_pos(io) # :nodoc:
        @positions[io] = io.pos if io.respond_to?(:pos)
      rescue Errno::ESPIPE => e
      end
    end
  end
end

module UU
  module OS
    module HTTP

      # HTTPClient is used for execution of common HTTP requests.
      # @example Basic usage:
      #    client = UU::OS::HTTP::HTTPClient.new('https://api.plus4u.net', 'c00', 'uu/os/Artifact')
      #    get = client.prepare_get('getAttributes')
      #    get.headers[:authorization] = 'Basic ...'
      #    get.headers[:accept] = 'application/json'
      #    result = get.execute
      #    attributes_json = result.body
      class HTTPClient
        include UU::OS::Util::URLConverter

        # Creates new instance of wrapped HTTPClient
        # including default configuration.
        # @private
        def self.__new_client__ # :nodoc:
          client = ::HTTPClient.new
          client.agent_name = USER_AGENT
          client.cookie_manager = nil
          client.request_filter.clear
          client.force_basic_auth = false
          client.instance_variable_set(:@proxy_auth, nil)
          client.instance_variable_set(:@www_auth, nil)
          client.follow_redirect_count = 0
          client.connect_timeout = 1
          client.send_timeout = 30
          client.receive_timeout = 30
          # enabling also forces setting of accept-encoding header
          client.transparent_gzip_decompression = DEFAULT_GZIP_DECOMPRESSION
          # configure SSL
          client.ssl_config.verify_mode = DEFAULT_SSL_VALIDATION
          client.ssl_config.ssl_version = DEFAULT_SSL_VERSION
          client.ssl_config.cert_store = DEFAULT_CERT_STORE
          # configure keepalive
          client.keep_alive_timeout = DEFAULT_KEEPALIVE_TIMEOUT
          return client
        end
        private_class_method :__new_client__

        # Default configuration of transparent GZIP decompression
        # @private
        DEFAULT_GZIP_DECOMPRESSION = true # :nodoc:
        private_constant :DEFAULT_GZIP_DECOMPRESSION

        # Configuration value for enabling SSL validation.
        # @private
        SSL_VALIDATION_ENABLED = OpenSSL::SSL::VERIFY_PEER | OpenSSL::SSL::VERIFY_FAIL_IF_NO_PEER_CERT # :nodoc:
        private_constant :SSL_VALIDATION_ENABLED

        # Configuration value for disabling SSL validation.
        # @private
        SSL_VALIDATION_DISABLED = OpenSSL::SSL::VERIFY_NONE # :nodoc:
        private_constant :SSL_VALIDATION_DISABLED

        # Default SSL validation configuration.
        # @private
        DEFAULT_SSL_VALIDATION = SSL_VALIDATION_ENABLED # :nodoc:
        private_constant :DEFAULT_SSL_VALIDATION

        # Default SSL version configuration.
        # @private
        DEFAULT_SSL_VERSION = :auto # :nodoc:
        private_constant :DEFAULT_SSL_VERSION

        # Default SSL certificate store.
        # @private
        DEFAULT_CERT_STORE = OpenSSL::X509::Store.new # :nodoc:
        DEFAULT_CERT_STORE.add_file(File.join(File.dirname(__FILE__), 'cacert.crt'))

        # Default keepalive timeout in seconds (15 is from Apache 2 default).
        # @private
        DEFAULT_KEEPALIVE_TIMEOUT = 15 # :nodoc:
        private_constant :DEFAULT_KEEPALIVE_TIMEOUT

        # User-Agent HTTP header.
        # @private
        USER_AGENT = "#{self} (#{UUOSConnection::VERSION}; #{RUBY_ENGINE}-" # :nodoc:
        if RUBY_ENGINE == 'jruby'
          require 'java'
          USER_AGENT << "#{JRUBY_VERSION}/#{RUBY_VERSION}; #{java.lang.System.getProperty('java.version')}; #{java.lang.System.getProperty('java.vm.name')}"
        else
          USER_AGENT << "#{RUBY_VERSION}"
        end
        USER_AGENT << "; #{RbConfig::CONFIG['target_os']}; #{RbConfig::CONFIG['host_cpu']})"
        private_constant :USER_AGENT

        # Default instance of HTTPClient. It is thread-safe and therefore
        # globally shared to utilize connection pool as much as possible.
        # @private
        DEFAULT_CLIENT = __new_client__ # :nodoc:
        private_constant :DEFAULT_CLIENT

        # Default set of HTTP headers
        # @private
        DEFAULT_HEADERS = UU::OS::HTTP::HTTPHeaders.new(content_type: 'application/octet-stream', accept: '*/*', user_agent: USER_AGENT) # :nodoc:
        private_constant :DEFAULT_HEADERS

        # Creates new instance of {HTTPClient HTTPClient} for given base URL
        # and optional path.
        # @param [String] base_url Base URL to connect in format +protocol+://+host+[:+port+].
        # @param [Array<String>] path Optional path to be appended to base URL.
        def initialize(base_url, *path)
          @base_url = base_url
          @path = concat_path(*path)
          @client = nil
        end

        # Returns timeout for connection initialization.
        # @return [Fixnum, Float] Connection timeout in seconds.
        def connect_timeout
          client.connect_timeout
        end

        # Sets timeout for connection initialization.
        # @param [Fixnum, Float] timeout Connection timeout in seconds.
        # @return [void]
        def connect_timeout=(timeout)
          return if timeout.nil?
          # Do not create new instance if default value is not changed
          if client.connect_timeout != timeout
            @client = self.class.send(:__new_client__) if @client.nil?
            @client.connect_timeout = timeout
          end
        end

        # Returns timeout in which request must be processed
        # (sum of sending of request and receiving response).
        # @return [Fixnum, Float] Request timeout in seconds.
        def request_timeout
          client.receive_timeout
        end

        # Sets timeout in which request must be processed
        # (sum of sending of request and receiving response).
        # @param [Fixnum, Float] timeout Request timeout in seconds.
        # @return [void]
        def request_timeout=(timeout)
          return if timeout.nil?
          # Do not create new instance if default value is not changed
          if client.receive_timeout != timeout
            @client = self.class.send(:__new_client__) if @client.nil?
            @client.receive_timeout = timeout
            @client.send_timeout = timeout
          end
        end

        # Returns configuration of transparent GZip decompression.
        # @return [TrueClass, FalseClass] Actual configuration.
        def tranparent_gzip_decompression
          client.transparent_gzip_decompression
        end

        # Enable or disable transparent GZip decompression. If enabled
        # corresponding Accept header is automatically send with request.
        # @param [TrueClass, FalseClass] enabled True to enable, False to disable.
        # @return [void]
        def tranparent_gzip_decompression=(enabled)
          return if enabled.nil?
          # Do not create new instance if default value is not changed
          if client.gransparent_gzip_decompression != enabled
            @client = self.class.send(:__new_client__) if @client.nil?
            @client.transparent_gzip_decompression = enabled
          end
        end

        # Mainly for testing purposes it might be necessary to override
        # default configuration of transparent GZip decompression.
        # @param [TrueClass, FalseClass] enabled True to enable, False to disable.
        # @return [void]
        # @private
        def self.transparent_gzip_decompression=(enabled)
          return if enabled.nil?
          self.send(:remove_const, :DEFAULT_GZIP_DECOMPRESSION) if self.const_defined?:DEFAULT_GZIP_DECOMPRESSION
          self.const_set(:DEFAULT_GZIP_DECOMPRESSION, enabled)
          DEFAULT_CLIENT.transparent_gzip_decompression = DEFAULT_GZIP_DECOMPRESSION
        end

        # Returns configuration of SSL validation.
        # @return [TrueClass, FalseClass] Actual configuration.
        def ssl_validation
          return client.ssl_config.verify_mode == SSL_VALIDATION_ENABLED
        end

        # Enable or disable SSL validation. If enabled, server certificate is checked
        # and exception is thrown in case it is not valid or is untrusted.
        # @param [TrueClass, FalseClass] enabled True to enable, False to disable.
        # @return [void]
        def ssl_validation=(enabled)
          return if enabled.nil?
          # Do not create new instance if default value is not changed
          if (client.ssl_config.verify_mode == SSL_VALIDATION_ENABLED) != enabled
            @client = self.class.send(:__new_client__) if @client.nil?
            if enabled
              client.ssl_config.verify_mode = SSL_VALIDATION_ENABLED
            else
              client.ssl_config.verify_mode = SSL_VALIDATION_DISABLED
            end
          end
        end

        # Mainly for testing purposes it might be necessary to override
        # SSL configuration on global client instance, so that new instance
        # is not created for every use (which will effectively disable
        # connection keep alive resulting in worse performance).
        # @param [TrueClass, FalseClass] enabled True to enable, False to disable.
        # @return [void]
        # @private
        def self.ssl_validation=(enabled) # :nodoc:
          return if enabled.nil?
          self.send(:remove_const, :DEFAULT_SSL_VALIDATION) if self.const_defined?:DEFAULT_SSL_VALIDATION
          if enabled
            self.const_set(:DEFAULT_SSL_VALIDATION, SSL_VALIDATION_ENABLED)
          else
            self.const_set(:DEFAULT_SSL_VALIDATION, SSL_VALIDATION_DISABLED)
          end
          DEFAULT_CLIENT.ssl_config.verify_mode = DEFAULT_SSL_VALIDATION
        end

        # Returns SSL version to be used for communication. Value :auto or
        # +nil+ means suitable version is automatically chosen based on
        # negotiation with server.
        # @return [String, Symbol] SSL version
        def ssl_version
          return client.ssl_config.ssl_version
        end

        # Sets SSL version to be used for communication.
        # @param [String, Symbol] version Possible values are TLSv1_2, TLSv1_1, TLSv1,
        #    SSLv2, SSLv23, SSLv3 or :auto (or +nil+) to allow version negotiation.
        # @return [void]
        def ssl_version=(version)
          # Do not create new instance if default value is not changed
          if (((client.ssl_config.ssl_version.nil?) || (client.ssl_config.ssl_version == :auto)) && !((version.nil?) || (version == :auto)))
            @client = self.class.send(:__new_client__) if @client.nil?
            @client.ssl_config.ssl_version = (version.nil? ? :auto : version)
          end
        end

        # Mainly for testing purposes it might be necessary to override
        # SSL configuration on global client instance, so that new instance
        # is not created for every use (which will effectively disable
        # connection keep alive resulting in worse performance).
        # @param [String, Symbol] version Possible values are TLSv1_2, TLSv1_1, TLSv1,
        #    SSLv2, SSLv23, SSLv3 or :auto (or +nil+) to allow version negotiation.
        # @return [void]
        # @private
        def self.ssl_version=(version) # :nodoc:
          self.send(:remove_const, :DEFAULT_SSL_VERSION) if self.const_defined?:DEFAULT_SSL_VERSION
          self.const_set(:DEFAULT_SSL_VERSION, (version.nil? ? :auto : version))
          DEFAULT_CLIENT.ssl_config.ssl_version = DEFAULT_SSL_VERSION
        end

        # Sets custom SSL certificate store to be able to modify default trusted CAs.
        # @param [String] cert_store Path to SSL certificate store.
        #    Path may point either to single file or directory containing
        #    multiple certificate files. PEM encoding is expected.
        # @return [void]
        def ssl_cert_store=(cert_store)
          return if cert_store.nil?
          @client = self.class.send(:__new_client__) if @client.nil?
          store = OpenSSL::X509::Store.new
          if File.directory?(cert_store)
            store.add_path(cert_store)
          else
            store.add_file(cert_store)
          end
          client.ssl_config.cert_store = store
        end

        # Mainly for testing purposes it might be necessary to override
        # SSL certificate store on global client instance, so that new
        # instance if not created for every use (which will effectively
        # disable connection keep alive resulting in worse performance).
        # @param [String] cert_store Path to SSL certificate store.
        #    Path may point either to single file or directory containing
        #    multiple certificate files. PEM encoding is expected.
        # @return [void]
        # @private
        def self.ssl_cert_store=(cert_store) # :nodoc:
          return if cert_store.nil?
          store = OpenSSL::X509::Store.new
          if File.directory?(cert_store)
            store.add_path(cert_store)
          else
            store.add_file(cert_store)
          end
          DEFAULT_CLIENT.ssl_config.cert_store = store
        end

        # Sets connection keepalive timeout (timeout within which are
        # connections to same server reused). Defaults to 15 seconds.
        # @param [Fixnum] timeout Keepalive timeout (in seconds).
        # @return [void]
        def keepalive_timeout=(timeout)
          return if timeout.nil?
          # Do not create new instance if default value is not changed
          if client.keep_alive_timeout != timeout.to_i
            @client = self.class.send(:__new_client__) if @client.nil?
            @client.keep_alive_timeout = timeout.to_i
          end
        end

        # Mainly for testing purposes it might be necessary to override
        # default keepalive timeout.
        # @param [Fixnum] timeout Keepalive timeout (in seconds).
        # @return [void]
        # @private
        def self.keepalive_timeout=(timeout) # :nodoc:
          self.send(:remove_const, :DEFAULT_KEEPALIVE_TIMEOUT) if self.const_defined?:DEFAULT_KEEPALIVE_TIMEOUT
          self.const_set(:DEFAULT_KEEPALIVE_TIMEOUT, (timeout.nil? ? 15 : timeout.to_i))
          DEFAULT_CLIENT.keep_alive_timeout = DEFAULT_KEEPALIVE_TIMEOUT
        end

        # Sets HTTP header to be used for each invoked request using this instance
        # of {HTTPClient HTTPClient}.
        # @param [Symbol, String] name Header name (case insensitive).
        # @param [Array<String>] value Header value (multiple values can be given).
        # @return [void]
        def set_header(name, *value)
          @headers = UU::OS::HTTP::HTTPHeaders.new(DEFAULT_HEADERS) if @headers.nil?
          @headers[name] = value
        end

        # Prepares GET request for given optional path. In case path is not given
        # request will be sent to URL defined by +base_url+ and +path+ given for
        # {HTTPClient HTTPClient} initialization.
        # @param [Array<String>] path Optional path where to send request (value is appended to +base_url+ and +path+ given for {#initialize HTTPClient initialization}).
        # @return [UU::OS::HTTP::HTTPGet] Instance of GET method.
        def prepare_get(*path)
          UU::OS::HTTP::HTTPGet.new(self, concat_path(*path))
        end

        # Prepares POST request for given optional path. In case path is not given
        # request will be sent to URL defined by +base_url+ and +path+ given for
        # {HTTPClient HTTPClient} initialization.
        # @param [Array<String>] path Optional path where to send request (value is appended to +base_url+ and +path+ given for {#initialize HTTPClient initialization}).
        # @return [UU::OS::HTTP::HTTPPost] Instance of POST method.
        def prepare_post(*path)
          UU::OS::HTTP::HTTPPost.new(self, concat_path(*path))
        end

        # Invokes request according to given parameters. In case you need to stream response or upload large data, {#request_async request_async} should be used.
        # @param [Symbol, String] method HTTP method (GET, POST, DELETE, etc.; case insensitive)
        # @param [String] path Path where to send request (value is appended to +base_url+ and +path+ given for {#initialize HTTPClient initialization}).
        # @param [Hash, UU::OS::HTTP::HTTPQuery] query Request query.
        # @param [String, IO] body Body to be sent (should match defined Content-Type).
        # @param [Hash, UU::OS::HTTP::HTTPHeaders] headers Additional request headers (to be merged with headers set using {#set_header set_header} method).
        # @return [UU::OS::HTTP::HTTPResponse] Response object.
        # @private Low level API wrapped by particular HTTP method object
        def request(method, path, query = nil, body = nil, headers = nil) # :nodoc:
          url = concat_path(@base_url, @path, path)
          merged_headers = headers().merge(headers)
          raw_request = prepare_request_data(method, query, body, merged_headers)
          raw_response = client.request(method, url, raw_request)
          request = UU::OS::HTTP::HTTPRequest.new(url, method, merged_headers, query, body)
          return UU::OS::HTTP::HTTPResponse.new(raw_response, request)
        end

        # Invokes asynchronous request according to given parameters. This method need to be used to stream response or upload large data.
        # @param [Symbol, String] method HTTP method (GET, POST, DELETE, etc.; case insensitive)
        # @param [String] path Path where to send request (value is appended to +base_url+ and +path+ given for {#initialize HTTPClient initialization}).
        # @param [Hash, UU::OS::HTTP::HTTPQuery] query Request query.
        # @param [String, IO] body Body to be sent (should match defined Content-Type).
        # @param [Hash, UU::OS::HTTP::HTTPHeaders] headers Additional request headers (to be merged with headers set using {#set_header set_header} method).
        # @return [UU::OS::HTTP::HTTPResponse] Response object.
        # @private Low level API wrapped by particular HTTP method object
        def request_async(method, path, query = nil, body = nil, headers = nil) # :nodoc:
          url = concat_path(@base_url, @path, path)
          merged_headers = headers().merge(headers)
          raw_request = prepare_request_data(method, query, body, merged_headers)
          raw_response = client.request_async2(method, url, raw_request)
          request = UU::OS::HTTP::HTTPRequest.new(url, method, merged_headers, query, body)
          return UU::OS::HTTP::HTTPResponse.new(raw_response, request)
        end

        # Override inspect to hide internal structures.
        # @private
        def inspect # :nodoc:
          "#<#{self.class.name}:0x#{(self.object_id << 1).to_s(16)}>"
        end

        private

        # Returns global instance of wrapped HTTPClient or re-configured
        # local instance of wrapped HTTPClient.
        # @return [::HTTPClient] Instance of HTTPClient.
        # @private
        def client # :nodoc:
          @client || DEFAULT_CLIENT
        end

        # Returns default set of HTTP headers or custom set
        # defined by this instance.
        # @return [UU::OS::CON::HTTP::HTTPHeaders] Headers to be used.
        # @private
        def headers # :nodoc:
          @headers || DEFAULT_HEADERS
        end

        # Allow to invoke other HTTP methods than POST and GET
        # (needed e.g. for UDS which uses DELETE, but might be
        # necessary for integration with thirdparty services).
        # @private
        def method_missing(method, *args, &block) #:nodoc:
          method_name = method.to_s
          if method_name =~ /^prepare_/
            method_type = method_name[8..-1].to_sym
            UU::OS::HTTP::HTTPMethod.new(self, concat_path(*args), method_type)
          else
            super
          end
        end

        # Prepares request so that it is usable by wrapped HTTPClient.
        # @param [Symbol] method HTTP method (must be lowercase).
        # @param [Hash, UU::OS::HTTP::HTTPQuery] query Request query.
        # @param [String, IO] body Body to be sent.
        # @param [Hash, UU::OS::HTTP::HTTPHeaders] headers Additional request headers.
        # @return [Hash] Transformed request data.
        # @private
        def prepare_request_data(method, query, body, headers) # :nodoc:
          request_data = {header: nil, query: nil, body: body}
          if (body.nil? && ((method == :post) || method == :put))
            # HTTPClient does not send content-length for null body, but
            # it might be required by some web servers (e.g. by WEBrick
            # which is used for testing)
            headers[:content_length] = 0
          end

          transformed_headers = transform_headers(headers)
          transformed_query = transform_query(query)

          request_data[:header] = transformed_headers if (!transformed_headers.nil?) && (!transformed_headers.empty?)
          request_data[:query] = transformed_query if (!transformed_query.nil?) && (!transformed_query.empty?)

          return request_data
        end

        # Transforms Hash to array of 2 element key-value arrays, where
        # all underscores are replaced by dashes in keys and entries are
        # omitted in case value is +nil+ (this is preferred form of
        # parameters for HTTPClient).
        # @param [Hash] data Hash to be transformed
        # @return [Array<Array<String>>] Hash transformed to Array
        # @private
        def transform_headers(data) # :nodoc:
          result = []
          data.each do |key, value|
            key = key.to_s.gsub('_', '-')
            if value.respond_to?:each
              value.each do |value|
                result << [key, value] if !value.nil?
              end
            else
              result << [key, value] if !value.nil?
            end
          end
          result
        end

        # Transforms Hash to array of 2 element key-value arrays,
        # Hash entry is omitted in case value is +nil+ (this is
        # preferred form of parameters for HTTPClient).
        # @param [Hash] data Hash to be transformed
        # @return [Array<Array<String>>] Hash transformed to Array
        # @private
        def transform_query(data) # :nodoc:
          result = []
          data.each do |key, value|
            key = key.to_s
            if value.respond_to?:each
              value.each do |value|
                result << [key, value] if !value.nil?
              end
            else
              result << [key, value] if !value.nil?
            end
          end
          result
        end

      end

    end
  end
end
